/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fmlmem.c,v $
 * Revision 1.13  1996/01/29 17:52:05  adam
 * Minor change - removed unnecessary &.
 *
 * Revision 1.12  1995/05/16  09:39:34  adam
 * LICENSE.
 *
 * Revision 1.11  1995/03/02  10:18:48  adam
 * Bug fix.
 *
 * Revision 1.10  1995/03/02  08:06:03  adam
 * Fml function strsub implemented. New test files marc[45].fml.
 * New test options in fmltest.
 *
 * Revision 1.9  1995/02/27  09:01:20  adam
 * Regular expression support. Argument passing by name option. New FML
 * function strlen.
 *
 * Revision 1.8  1995/02/23  08:32:05  adam
 * Changed header.
 *
 * Revision 1.6  1995/02/10  18:15:52  adam
 * FML function 'strcmp' implemented. This function can be used to
 * test for existence of MARC fields.
 *
 * Revision 1.5  1995/02/09  14:37:18  adam
 * Removed .depend from cvs. Removed function fml_mk_list.
 *
 * Revision 1.4  1995/02/09  14:33:37  adam
 * Split source fml.c and define relevant build-in functions in separate
 * files. New operators mult, div, not, llen implemented.
 *
 * Revision 1.3  1995/02/09  13:07:15  adam
 * Nodes are freed now. Many bugs fixed.
 *
 * Revision 1.2  1995/02/06  15:23:26  adam
 * Added some more relational operators (le,ne,ge). Added increment
 * and decrement operators. Function index changed, so that first
 * element is 1 - not 0. Function fml_atom_val edited.
 *
 * Revision 1.1.1.1  1995/02/06  13:48:10  adam
 * First version of the FML interpreter. It's slow and memory isn't
 * freed properly. In particular, the FML nodes aren't released yet.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "fmlp.h"

#define FML_ATOM_CHUNK 1024
#define FML_NODE_CHUNK 1024

static int no_nodes = 0;
static int no_atoms = 0;

struct fml_node *fml_node_alloc (Fml fml)
{
    struct fml_node *n;

    if (! fml->node_free_list)
    {
        int i;

        n = fml->node_free_list = malloc (sizeof(*n) * FML_NODE_CHUNK);
        if (!n)
        {
            (*fml->err_handle)(FML_ERR_NOMEM);
            exit (1);
        }
        for (i = FML_ATOM_CHUNK-1; --i >= 0; n++)
            n->p[1] = n+1;
        n->p[1] = NULL;
    }
    n = fml->node_free_list;
    fml->node_free_list = n->p[1];
    n->p[0] = n->p[1] = NULL;
    n->is_atom = 0;
    no_nodes++;
    return n;
}

static struct fml_atom *atom_malloc (Fml fml)
{
    struct fml_atom *fa;

    if (! fml->atom_free_list)
    {
        int i;

        fa = fml->atom_free_list = malloc (sizeof(*fa) * FML_ATOM_CHUNK);
        if (!fa)
        {
            (*fml->err_handle)(FML_ERR_NOMEM);
            exit (1);
        }
        for (i = FML_ATOM_CHUNK-1; --i >= 0; fa++)
            fa->next = fa+1;
        fa->next = NULL;
    }
    fa = fml->atom_free_list;
    fml->atom_free_list = fa->next;
    no_atoms++;
    return fa;
}

static void atom_delete (Fml fml, struct fml_atom *a)
{
    a->next = fml->atom_free_list;
    fml->atom_free_list = a;
    no_atoms--;
}

static struct fml_atom *atom_copy (Fml fml, struct fml_atom *a)
{
    struct fml_atom *a0, *a1;

    a0 = a1 = atom_malloc (fml);
    while (a)
    {
        memcpy (a1->buf, a->buf, FML_ATOM_BUF);
        if (!a->next)
            break;
        a = a->next;
        a1 = a1->next = atom_malloc (fml);
    }
    a1->next = NULL;
    return a0;
}

struct fml_atom *fml_atom_alloc (Fml fml, char *str)
{
    int soff = 0;
    struct fml_atom *a, *a0;

    a0 = a = atom_malloc (fml);
    strncpy (a->buf, str, FML_ATOM_BUF);
    while (strlen (str+soff) >= FML_ATOM_BUF)
    {
        struct fml_atom *an;

        an = atom_malloc (fml);
        a->next = an;
        soff += FML_ATOM_BUF;
        strncpy (an->buf, str+soff, FML_ATOM_BUF);
        a = an;
    }
    a->next = NULL;
    return a0;
}

int fml_atom_cmp (Fml fml, struct fml_atom *a1, struct fml_atom *a2)
{
    while (a1 && a2)
    {
        int n;
        n = strncmp (a1->buf, a2->buf, FML_ATOM_BUF);
        if (n)
            return n;
        a1 = a1->next;
        a2 = a2->next;
    }
    if (!a1 && !a2)
        return 0;
    if (a1)
        return 1;
    return -1;
}

int fml_atom_str (struct fml_atom *a, char *str)
{
    int i, len = 0;

    assert (a);
    while (a->next)
    {
        if (str)
            memcpy (str+len, a->buf, FML_ATOM_BUF);
        len += FML_ATOM_BUF;
        a = a->next;
    }
    for (i=0; i<FML_ATOM_BUF && a->buf[i]; i++)
        ;
    if (str)
    {
        memcpy (str+len, a->buf, i);
        str[len+i] = 0;
    }
    len += i;
    return len;
}

void fml_atom_strx (struct fml_atom *a, char *str, int max)
{
    int len = 0;

    assert (a);
    while (a->next && len < max - 2*FML_ATOM_BUF)
    {
        memcpy (str+len, a->buf, FML_ATOM_BUF);
        len += FML_ATOM_BUF;
        a = a->next;
    }
    strncpy (str+len, a->buf, FML_ATOM_BUF-1);
    str[len+FML_ATOM_BUF-1] = '\0';
}

int fml_atom_len (struct fml_atom *a)
{
    int len = 0;
    if (a)
    {
        while (a->next)
        {
            len += FML_ATOM_BUF;
            a = a->next;
        }
        len += strlen (a->buf);
    } 
    return len;
}

int fml_atom_val (struct fml_atom *a)
{
    static char arg[256];
    assert (a);
    if (!a->next)
        return atoi (a->buf);
    fml_atom_strx (a, arg, 200);
    return atoi (arg);
}

struct fml_node *fml_mk_node_val (Fml fml, int val)
{
    static char arg[64];
    struct fml_node *fn;

    sprintf (arg, "%d", val);
    fn = fml_node_alloc (fml);
    fn->is_atom = 1;
    fn->p[0] = fml_atom_alloc (fml, arg);
    return fn;
}

void fml_node_delete (Fml fml, struct fml_node *fn)
{
    struct fml_node *f1;
    while (fn)
    {
        if (fn->is_atom)
            atom_delete (fml, fn->p[0]);
        else
            fml_node_delete (fml, fn->p[0]);
        f1 = fn->p[1];
        
        fn->p[1] = fml->node_free_list;
        fml->node_free_list = fn;
        no_nodes--;

        fn = f1;
    }
}

struct fml_node *fml_node_copy (Fml fml, struct fml_node *fn)
{
    struct fml_node *fn0, *fn1;

    if (!fn)
        return NULL;
    fn1 = fn0 = fml_node_alloc (fml);
    while (1)
    {
        if (fn->is_atom)
        {
            fn1->is_atom = 1;
            fn1->p[0] = atom_copy (fml, fn->p[0]);
        }
        else 
            fn1->p[0] = fml_node_copy (fml, fn->p[0]);
        if (!fn->p[1])
            break;
        fn = fn->p[1];
        fn1 = fn1->p[1] = fml_node_alloc (fml);
    }
    return fn0;
}

void fml_node_stat (Fml fml)
{
    if (fml->debug & 2)
        printf ("<<node=%d, atom=%d>>", no_nodes, no_atoms);
}

struct fml_atom *fml_atom_strsub (Fml fml, struct fml_atom *a, int o, int l)
{
    static char buf[512];
    char *cp;
    struct fml_atom *an;
    int ol = fml_atom_len (a);

    if (ol >= 510)
    {
        cp = malloc (ol + 1);
        assert (cp);
    }
    else
        cp = buf;
    fml_atom_str (a, buf);
    if (o + l < ol)
        buf[o+l] = '\0';
    an = fml_atom_alloc (fml, buf+o);
    if (ol >= 510)
        free (cp);
    return an;
}
