/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fml.h,v $
 * Revision 1.5  1995/05/16 09:39:38  adam
 * LICENSE.
 *
 * Revision 1.4  1995/02/23  08:32:11  adam
 * Changed header.
 *
 * Revision 1.2  1995/02/22  08:51:02  adam
 * Output function can be customized.
 *
 * Revision 1.1  1995/02/20  21:15:46  adam
 * Moved files to include.
 *
 * Revision 1.3  1995/02/10  15:50:55  adam
 * MARC interface implemented. Minor bugs fixed. fmltest can
 * be used to format single MARC records. New function '\list'
 * implemented.
 *
 * Revision 1.2  1995/02/09  16:06:06  adam
 * FML can be called from the outside multiple times by the functions:
 * fml_exec_call and fml_exec_call_str.
 * An interactive parameter (-i) to fmltest starts a shell-like
 * interface to FML by using the fml_exec_call_str function.
 *
 * Revision 1.1.1.1  1995/02/06  13:48:10  adam
 * First version of the FML interpreter. It's slow and memory isn't
 * freed properly. In particular, the FML nodes aren't released yet.
 *
 */

#ifndef FML_H
#define FML_H
typedef struct Fml_record {
    struct fml_node *list;
    struct fml_sym_tab *sym_tab;

    struct fml_atom *atom_free_list;
    struct fml_node *node_free_list;

    int escape_char;
    int eof_mark;
    char *white_chars;
    char comment_char;
    int (*read_func)(void);
    void (*write_func)(int ch);
    void (*err_handle)(int no);

    int debug;
} *Fml;

Fml fml_open (void);
int fml_preprocess (Fml fml);

void fml_exec (Fml fml);
void fml_exec_call (Fml fml);
void fml_exec_call_str (Fml fml, const char *str);
void fml_exec_call_argv (Fml fml, const char **argv);

#define FML_ERR_NOMEM 1

#endif
