/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * Iso2709 record management - anchor utilities
 *
 * Europagate, 1995.
 *
 * $Log: iso2709a.c,v $
 * Revision 1.4  1995/05/16 09:40:54  adam
 * LICENSE.
 *
 * Revision 1.3  1995/03/30  14:22:18  adam
 * More work on new MARC anchor functions.
 *
 * Revision 1.2  1995/03/30  07:33:35  adam
 * New 2709 function: iso2709_mk.
 * First implementation of iso2709_a_insert.
 *
 * Revision 1.1  1995/03/29  11:44:29  adam
 * New functions: iso2709_a_.. for record manipulation.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include <iso2709p.h>

Iso2709Anchor iso2709_a_mk (Iso2709Rec rec)
{
    Iso2709Anchor anchor;
    
    anchor = malloc (sizeof(*anchor));
    if (!anchor)
        return NULL;
    anchor->rec = rec;
    anchor->d0 = &rec->directory;
    if (*anchor->d0)
        anchor->f0 = &(*anchor->d0)->fields;
    else
        anchor->f0 = NULL;
    return anchor;
}

void iso2709_a_rm (Iso2709Anchor anchor)
{
    free (anchor);
}

int iso2709_a_first (Iso2709Anchor anchor)
{
    anchor->d0 = &anchor->rec->directory;
    if (*anchor->d0)
    {
        anchor->f0 = &(*anchor->d0)->fields;
        return 1;
    }
    return 0;
}

int iso2709_a_next_line (Iso2709Anchor anchor)
{
    if (! *anchor->d0)
        return 0;
    anchor->d0 = &(*anchor->d0)->next;
    if (*anchor->d0)
        anchor->f0 = &(*anchor->d0)->fields;
    return 1;
}

int iso2709_a_next_field (Iso2709Anchor anchor)
{
    if (!*anchor->d0 || !*anchor->f0)
        return 0;
    if (!(*anchor->f0)->next)
        return 0;
    anchor->f0 = &(*anchor->f0)->next;
    return 1;
}

int iso2709_a_next (Iso2709Anchor anchor)
{
    if (!*anchor->d0 || !*anchor->f0)
        return iso2709_a_next_line (anchor);
    anchor->f0 = &(*anchor->f0)->next;
    if (! *anchor->f0)
        return iso2709_a_next_line (anchor);
    return 2;
}

int iso2709_a_info_field (Iso2709Anchor anchor,
                    char **tag, char **indicator,
                    char **identifier, char **data)
{
    if (!*anchor->d0 || !*anchor->f0)
        return 0;
    if (tag)
        *tag = (*anchor->d0)->tag;
    if (indicator)
    {
        *indicator = (*anchor->d0)->indicator;
        if (*indicator && !**indicator)
            *indicator = NULL;
    }
    if (identifier)
    {
        *identifier = (*anchor->f0)->identifier;
        if (*identifier && !**identifier)
            *identifier = NULL;
    }
    if (data)
        *data = (*anchor->f0)->data;
    return 1;
}

int iso2709_a_info_line (Iso2709Anchor anchor,
                    char **tag, char **indicator)
{
    if (!*anchor->d0)
        return 0;
    assert (*anchor->f0);
    return iso2709_a_info_field (anchor, tag, indicator, NULL, NULL);
}

int iso2709_a_delete_field (Iso2709Anchor anchor)
{
    struct iso2709_field *field;

    if (!*anchor->d0)
        return 0;
    field = *anchor->f0;
    *anchor->f0 = field->next;
    free (field->data);
    free (field);
    if (!*anchor->f0)
    {
        if (! (*anchor->d0)->fields)
            iso2709_a_delete_line (anchor);
        iso2709_a_next_line (anchor);
    }
    return 1;
}

int iso2709_a_delete_line (Iso2709Anchor anchor)
{
    struct iso2709_dir *dir;

    if (!*anchor->d0)
        return 0;
    dir = *anchor->d0;
    *anchor->d0 = dir->next;
    free (dir->indicator);
    free (dir);
    return 1;
}

static int marc_cmp (const char *field, const char *pattern)
{
    if (*pattern == '*')
        return 0;
    if (!field)
        return -*pattern;
    for (; *field && *pattern; field++, pattern++)
    {
        if (*pattern == '?')
            continue;
        if (*pattern != *field)
            break;
    }
    return *field - *pattern;
}

int iso2709_a_search (Iso2709Anchor anchor,
                      const char *tag_p, const char *indicator_p,
                      const char *identifier_p)
{
    char *tag;
    char *indicator;
    char *identifier;
    do
    {
        if (!iso2709_a_info_field (anchor, &tag, &indicator,
                                   &identifier, NULL))
            return 0;
        if ((!tag_p || !marc_cmp (tag, tag_p)) &&
            (!indicator_p || !marc_cmp (indicator, indicator_p)) &&
            (!identifier_p || !marc_cmp (identifier, identifier_p)))
            return 1;
    } while (iso2709_a_next (anchor));
    return 0;
}

int iso2709_a_insert (Iso2709Anchor a, 
                      const char *tag, const char *indicator,
                      const char *identifier, const char *data)
{
    struct iso2709_field *field;

    assert (data);
    assert (tag);
    assert (a->d0);
    if (! *a->d0 || strcmp (tag, (*a->d0)->tag))
    {
        struct iso2709_dir *dir;

        if (!(dir = malloc (sizeof(*dir))))
	    return 0;
        strcpy (dir->tag, tag);
	if (indicator)
	{
	    if (!(dir->indicator = malloc (strlen(indicator)+1)))
	        return 0;
            strcpy (dir->indicator, indicator);
	}
        else
	    dir->indicator = NULL;
        if (*a->d0)
            a->d0 = &(*a->d0)->next;
        dir->next = *a->d0;
        *a->d0 = dir;
	dir->fields = NULL;
	a->f0 = &dir->fields;
    }
    if (!(field = malloc (sizeof(*field))))
        return 0;
    if (identifier)
	strcpy (field->identifier, identifier);
    else
        *field->identifier = '\0';
    if (!(field->data = malloc (strlen (data)+1)))
	return 0;
    strcpy (field->data, data);
    field->next = *a->f0;
    *a->f0 = field;
    a->f0 = &field->next;
    return 1;
}
